/*
 * ** PIR Private Directory Service prototype
 * ** Copyright (C) 2003 Alexander Iliev <iliev@nimbus.dartmouth.edu>
 * **
 * ** This program is free software; you can redistribute it and/or modify
 * ** it under the terms of the GNU General Public License as published by
 * ** the Free Software Foundation; either version 2 of the License, or
 * ** (at your option) any later version.
 * **
 * ** This program is distributed in the hope that it will be useful,
 * ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 * ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * ** GNU General Public License for more details.
 * **
 * ** You should have received a copy of the GNU General Public License
 * ** along with this program; if not, write to the Free Software
 * ** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * */

#include "io.h"

#include <string>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <common/consts.h>

#include "hostcall.h"



static const std::string IOROOT = "/priv-db";




//
// HostIO
//


void make_object_name (object_name_t & o_name, const std::string& name);
void make_object_name (object_name_t & o_name, const object_id& id);


HostIO::HostIO (const std::string & basedir) :
    _basedir (basedir)
{}


void HostIO::write (const object_id& id,
		    const ByteBuffer& obj)
    throw (hostio_exception)
{
    object_name_t name;
    make_object_name (name, id);

    try {
	host_write_blob (_basedir, name, obj);
    }
    catch (const better_exception & ex) {
	throw hostio_exception (ex.what());
    }
}


ByteBuffer HostIO::read (const object_id& id)
    throw (hostio_exception)
{
    ByteBuffer answer;
    object_name_t name;

    make_object_name (name, id);
    try {
	host_read_blob (_basedir, name, answer);
    }
    catch (const better_exception & ex) {
	throw hostio_exception (ex.what());
    }
    
    return answer;
}

//
// and with object naming using strings...
//

void HostIO::write (const std::string& id,
		    const ByteBuffer& obj)
    throw (hostio_exception)
{
    object_name_t name;
    make_object_name (name, id);

    try {
	host_write_blob (_basedir, name, obj);
    }
    catch (const better_exception & ex) {
	throw hostio_exception (ex.what());
    }
}


ByteBuffer HostIO::read (const std::string& id)
    throw (hostio_exception)
{
    ByteBuffer answer;
    object_name_t name;

    make_object_name (name, id);
    try {
	host_read_blob (_basedir, name, answer);
    }
    catch (const better_exception & ex) {
	throw hostio_exception (ex.what());
    }
    
    return answer;
}





void make_object_name (object_name_t & o_name, const std::string& name) {
    o_name.name_scheme              = NAME_STRING;
    o_name.object_name_t_u.str_name = const_cast<char*> (name.c_str());
}

void make_object_name (object_name_t & o_name, const object_id& id) {
    o_name.name_scheme              = NAME_OBJECT_ID;
    o_name.object_name_t_u.oid_name = id;
}



// these not necessary for now

#if 0

string utoa (unsigned u) {

    if (u == 0) return "0";

    string answer;
    while (u > 0) {
	answer.insert (0, 1, '0' + char(u % 10));
	u /= 10;
    }
    return answer;
}



static inline string getfilename (const string& basedir,
				  const object_id& id)
{
    return IOROOT + "/" + basedir + "/" +
	utoa (id.bucket) + "/" +
	utoa (id.record);
}

static inline string get_buckdir_name (const string& basedir,
				       const object_id& id)
{
    return IOROOT + "/" + basedir + "/" + utoa (id.bucket);
}

				  
void host_write_obj (const std::string& basedir,
		     const object_id& id,
		     const ByteBuffer& obj)
    throw (host_exception)
{
    int fd;
    int rc;
    string filename = getfilename (basedir, id);
    fd = open (filename.c_str(), O_WRONLY | O_TRUNC);
    if (fd == -1) {
	if (errno != ENOENT) {
	    // a bad error!
	    throw host_exception (string("host_write_obj open: ") +
				  strerror (errno));
	}

	rc = builddirs (filename, 00700);
	if (rc != 0) {
	    throw host_exception
		(string("host_write_obj making parent directories: ") +
		 strerror (errno));
	}

	fd = open (filename.c_str(), O_WRONLY | O_TRUNC);
    }
    
    rc = write (fd, obj.data(), obj.len());
    if (rc != obj.len()) {
	close (fd);
	throw host_exception (string("host_write_obj write: ")
			      + strerror (errno));
    }

    if (close (fd) == -1) {
	throw host_exception (string("host_write_obj close: ")
			      + strerror (errno));
    }
}



ByteBuffer host_read_obj (const std::string& basedir,
			  const object_id& id)
    throw (host_exception)
{
    int fd;
    int rc;
    ByteBuffer answer (new byte[BUFSIZE], BUFSIZE);
    
    string filename = getfilename (basedir, id);
    fd = open (filename.c_str(), O_RDONLY);
    if (fd == -1) {
	throw host_exception (string("host_read_obj open: ")
			      + strerror (errno));
    }
    
    rc = read (fd, answer.data(), answer.len());
    if (rc == -1) {
	close (fd);
	throw host_exception (string("host_read_obj read: ")
			      + strerror (errno));
    }
    else if (rc == answer.len()) {
	// uf, overflow, avoid handling for now
	close (fd);
	throw host_exception ("host_read_obj read: buffer too small!");
    }

    answer.len() = rc;

    if (close (fd) == -1) {
	throw host_exception (string("host_read_obj close: ")
			      + strerror (errno));
    }

    return answer;
}

#endif // 0
