/*
 * ** PIR Private Directory Service prototype
 * ** Copyright (C) 2003 Alexander Iliev <iliev@nimbus.dartmouth.edu>
 * **
 * ** This program is free software; you can redistribute it and/or modify
 * ** it under the terms of the GNU General Public License as published by
 * ** the Free Software Foundation; either version 2 of the License, or
 * ** (at your option) any later version.
 * **
 * ** This program is distributed in the hope that it will be useful,
 * ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 * ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * ** GNU General Public License for more details.
 * **
 * ** You should have received a copy of the GNU General Public License
 * ** along with this program; if not, write to the Free Software
 * ** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * */

#include <iostream>
#include <iomanip>
#include <vector>
#include <functional>

#include <stdlib.h>

#include <time.h>
#include <unistd.h>		// for sleep()

#include <common/utils.h>

#include "bit_matrix.h"
#include "permute.h"


using namespace std;



class simple_switch {

public:
    simple_switch (vector<index_t> & V) : V(V) {}
    
    void operator () (index_t a, index_t b,
		      bool should_swap);

private:
    vector<index_t> & V;
};



void simple_switch::operator () (index_t a, index_t b,
				 bool should_swap)
{
    if (should_swap) {
	swap (V[a], V[b]);
    }
}




void print_permutation (ostream & os, const perm_vector & V);

// P_start and P_stride are in wire numbers
void trace_perm (int level,
		 size_t P_size, size_t P_start, size_t P_stride,
		 size_t S_start_col, size_t S_start_row,
		 size_t S_cols, size_t S_rows);

string to_binary (unsigned int a, size_t bitsize);




void run_permutation
(
    vector<index_t> & V,	// what to permute
    const vector_stride_info_t & V_slice, // the parts of V to touch
    size_t N,
    const bit_matrix & S,	// the switches of the network
    const matrix_slice_info_t & S_slice	// which parts to use
    );




template<class T>
struct iterate_generator {
    typedef T result_type;

    iterate_generator (const unary_function<T,T>& f, const result_type& start)
	: f(f), val(start) {}

    result_type operator() ()
	{ result_type temp = val; val = f(val); return temp; }

    unary_function<T,T> f;
    result_type val;
};


template<class T>
iterate_generator<T> iterate (const unary_function<T,T>& f, const T& start) {
    return iterate_generator<T> (f, start);
}






#ifdef _TESTING_PERMUTE_C

int main (int argc, char * argv[]) {

    size_t N = atoi(argv[1]);

#if 0
    int l = atoi(argv[2]);

    size_t n = lgN_floor (N);


    for (index_t a=0; a < N; a++) {
	index_t r = revslide (a, l, n);
	cout << to_binary(a, n) << " (" << a << ") at l " << l << " -> "
	     << endl
	     << to_binary (r, n) << " (" << r << ")"
	     << endl << endl;
    }

    return 0;
#endif

    timeval tv;
    gettimeofday(&tv, NULL);

//    clog << tv.tv_usec << endl;
    
    srand ( argc > 2 ?
	    atoi(argv[2]) :
	    ( tv.tv_usec ) );
    
    perm_vector P;
    generate_permutation (P, N);
    print_permutation (cout, P);
    
    // allocate the switch settings bit matrix
    bit_vector S_template (N/2, false);
    bit_matrix S (2*lgN_floor(N) - 1, S_template);

    compute_switch_network (P, N, S);


    vector<index_t> V (N);
    // iterate (+ 1) over V, to fill it with [0..N-1]
//      generate (V.begin(), V.end(),
//  	      iterate (bind2nd (plus<index_t>(), 1), index_t(0)));
    for (index_t i=0; i < N; i++) {
	V[i] = i;
    }

    simple_switch sw (V);
    run_permutation_ct (N, S, sw);
    
//      run_permutation (V, vector_stride_info_t(),
//  		     N,
//  		     S, matrix_slice_info_t());

//    clog << "The result of permuting: " << endl;
    for (index_t i=0; i < N; i++) {
	cout << V[i] << " ";
    }
    cout << endl;
    
/*
    trace_perm (0,
		N, 0, 1,
		0, 0,
		2*lgN_floor(N) - 1, N/2);
*/
    
}

#endif // _TESTING_PERMUTE_C


void print_permutation (ostream & os, const perm_vector & V) {
    perm_vector::const_iterator it;
    int i;
    for (it = V.begin(), i = 0;
	 it != V.end();
	 it++, i++)
    {
	os << it->src << " ";
//  	os << setw(3) << i << " -> " << setw(3) << it->dest
//  	   << ", " << setw(3) << it->src
//  	   << endl;
    }
    os << endl;
}

    
void generate_permutation (perm_vector & V, size_t N) {

    V.resize(N);
    for (index_t i = 0; i < N; i++) {
	V[i].src = i;
    }

    // FIXME: reconsider the randomness source here
    
    for (index_t i = N-1; i > 0; i--) {
	int j = random () % (i+1);
	swap (V[i], V[j]);
    }

    for (index_t i = 0; i < N; i++) {
	V[V[i].src].dest = i;
    }
}


string to_binary (unsigned int a, size_t bitsize) {

    string answer = "";
    for (unsigned int i=0; i < bitsize; i++) {
	answer = ((a & 1) ? " 1" : " 0") + answer;
	a >>= 1;
    }

    return answer;
}



void print_switch (int s,
		   const string& label,
		   size_t P_size, size_t P_start, size_t P_stride,
		   size_t S_col, size_t S_start_row,
		   int level)
{
    int w1 = P_start + s * 2*P_stride;
    int w2 = w1 + P_stride;
    cout << level << label << ": switch wires " << setw(3) << w1 << " & "
	 << setw(3) << w2
	 << " with switch (" << setw(3) << S_col << ","
	 << to_binary(S_start_row + s, lgN_floor(P_size)-1) << ")"
	 << endl;
}


// P_start and P_stride are in wire numbers
void trace_perm (int level,
		 size_t P_size,
		 size_t P_start, size_t P_stride,
		 size_t S_start_col, size_t S_start_row,
		 size_t S_cols, size_t S_rows)
{

    size_t num_switches = P_size/(P_stride*2);
    
    
    if (num_switches == 1) {
	print_switch (0, " ",
		      P_size, P_start, P_stride,
		      S_start_col, S_start_row,
		      level);
	return;
    }

    // iterate over the switches! s is the switch number
    for (unsigned int s=0; s < num_switches; s++) {
	print_switch (s, "a",
		      P_size, P_start, P_stride,
		      S_start_col, S_start_row,
		      level);
    }

    trace_perm (level+1, P_size, P_start, P_stride*2,
		S_start_col+1, S_start_row,
		S_cols-2, S_rows/2);
	
    trace_perm (level+1, P_size, P_start+P_stride, P_stride*2,
		S_start_col+1, S_start_row + S_rows/2,
		S_cols-2, S_rows/2);

    
    // iterate over the switches! s is the switch number
    for (unsigned int s=0; s < num_switches; s++) {
	print_switch (s, "b",
		      P_size, P_start, P_stride,
		      S_start_col + (S_cols-1), S_start_row,
		      level);
    }

    
}


void add_to_sub_perm (perm_vector & P,
		      index_t left_idx, index_t right_idx)
{
    P[left_idx/2].dest = right_idx/2;
    P[right_idx/2].src = left_idx/2;
}



//
// a little

// compute the switch network S for the permutation P
// 
// S should be the right size now, with all switches set to false, ie do not
// cross.
void
compute_switch_network
(
    const perm_vector & P, // the permutation
    size_t N,			// number of wires
    bit_matrix & switch_matrix,	// output parameter: the switches
    const matrix_slice_info_t & slice // the slice information for S
    )
{

    //
    // a few utility macros for the criss-crossing loop
    //

    // which way does an index "want" to be routed: evens go up and odds go down
#define WANT_ROUTE(i) ( ((i)&1) == 0 ? UP : DOWN )

    // what is the index of this wire's switch partner? just flip the low bit
#define PARTNER(i) ( (i)^1 )

    // what is the switch number for a wire?
#define SWITCH_NUMBER(w) ( (w)/2 )

    enum { LEFT=0, RIGHT };	// names for the left and right switch columns

    enum subperm_t { UP=0, DOWN }; // names for the two sub-networks


//    clog << "****N = " << N << endl;
    
    //
    // base case, one switch only
    //
    if (N == 2) {
	if (P[0].dest != 0) {	// need to cross
//	    clog << "crossing single switch" << endl;
	    idx (switch_matrix, slice, 0, 0) = true;
	}
	return;
    }

    // flags for which switches have been looked at, only needed on the left
    // side
    bit_vector touched (N/2, false);

    // the two columns of switches of interest
    bit_matrix::reference
	S_left  = column (switch_matrix, slice, 0),
	S_right = column (switch_matrix, slice, 2*lgN_floor(N) - 2);
    
    
    perm_vector subperm[2];	// the sub-permutations for the two
				// sub-permuters
    subperm[UP].resize (N/2);
    subperm[DOWN].resize (N/2);
	
    // where to start the next round of criss-crossing?
    index_t start = 0;


    do {
	
	index_t left, right;	// wire indices for the two switch columns

	left = start * 2;

//	clog << "Start crisscross at left = " << left << endl;
	
	//
	// do this round of criss-crossing
	//
	do {
	    
	    // this switch is always straight
//  	    clog << "touching left switch " << SWITCH_NUMBER(left)
//  		 << ", wire " << PARTNER(left) << endl;
	    touched[SWITCH_NUMBER(left)] = true;
	    
	    right = P[left].dest; // where is the permuted value 'left'?

	    add_to_sub_perm (subperm[UP], left, right);

//  	    clog << "touching right switch " << SWITCH_NUMBER(right)
//  		 << ", wire " << right << endl;
	    if (WANT_ROUTE(right) != UP) {
		// need to cross the switch on the right!
//  		clog << "crossing right switch " << SWITCH_NUMBER(right)
//  		     << endl;
		idx (S_right, slice, SWITCH_NUMBER(right)) = true;
	    }
	    
	    // back to the left column
	    
	    right = PARTNER(right);
	    left = P[right].src; // where did value at 'right' come from?

	    add_to_sub_perm (subperm[DOWN], left, right);

	    // this switch setting is duplicated if the switch on the left was
	    // already set (ie at the end of a criss-cross)
	    if (WANT_ROUTE(left) != DOWN) {
		// need to cross this switch on the left
//  		clog << "crossing left switch " << SWITCH_NUMBER(left)
//  		     << endl;
		idx (S_left, slice, SWITCH_NUMBER(left)) = true;
	    }

	    left = PARTNER(left); // ready for next jump to the right

	} while (!touched[SWITCH_NUMBER(left)]);

	
	// find where to start from the next time round
	for (; start < N/2 && touched[start]; start++) ;

    } while (start < N/2);


    // and now do the sub-permuters!

//      clog << "Permuter UP" << endl; print_permutation (clog, subperm[UP]);
    compute_switch_network (subperm[UP], N/2,
			    switch_matrix, matrix_slice_info_t (slice, 1, 0));

//      clog << "Permuter DOWN" << endl; print_permutation (clog, subperm[DOWN]);
    compute_switch_network (subperm[DOWN], N/2,
			    switch_matrix, matrix_slice_info_t (slice, 1, N/4));
    
}



void print_vector (const string& msg, const vector<index_t> & V) {
    return;
    clog << msg << endl;
    for (index_t i=0; i < V.size(); i++) {
	clog << V[i] << " ";
    }
    clog << endl;
}    


void run_permutation
(
    vector<index_t> & V,	// what to permute
    const vector_stride_info_t & V_slice, // the parts of V to touch
    size_t N,
    const bit_matrix & S,		// the switches of the network
    const matrix_slice_info_t & S_slice	// which parts to use
    )

{
    if (N <= 2) {
	print_vector ("Vector before single switch", V);
	
	// base case: only one switch
	if (idx (S, S_slice, 0, 0)) {
	    swap ( idx (V, V_slice, 0),
		   idx (V, V_slice, 1) );
	}

 	print_vector ("Vector after single switch", V);

	return;
    }

    
    // do the first round of switches
    const bit_vector & S_left = column (S, S_slice, 0);
    for (index_t i=0; i < N; i+=2) { // i will be a wire number
	if (idx (S_left, S_slice, i/2)) { // is switch i/2 crossed?
	    swap ( idx (V, V_slice, i),
		   idx (V, V_slice, i+1) );
	}
    }

    print_vector ("Vector after first switch column:", V);
    
    // two recursions

    // evens of V, upper sub-permuter in S
    run_permutation (V,
		     vector_stride_info_t (V_slice, 0, 2),
		     N/2,
		     S,
		     matrix_slice_info_t (S_slice, 1, 0));

    // odds of V, lower sub-permuter in S
    run_permutation (V,
		     vector_stride_info_t (V_slice, 1, 2),
		     N/2,
		     S,
		     matrix_slice_info_t (S_slice, 1, N/4));


    print_vector ("Vector after subperms:", V);
    
    // the last round of switches
    const bit_vector & S_right = column (S, S_slice, 2*lgN_floor(N) - 2);
    for (index_t i=0; i < N; i+=2) { // i will be a wire number
	if (idx (S_right, S_slice, i/2)) { // is switch i/2 crossed?
	    swap ( idx (V, V_slice, i),
		   idx (V, V_slice, i+1) );
	}
    }

}



// one index operation for a revslide permutation:
// <---n-l-----><--l--->
// |     X     |   Y   | goes to
// | rev(Y)|      X    |
// l is the level number
// n is the bit size of the indices
index_t revslide (index_t a, size_t l, size_t n) {

    index_t r_X = a >> l;	// get the X part shifted right

#define GETBIT(a,i)   ( (a) & (1U << (i)) ) >> (i)
#define SETBIT(a,i,b) (a) |= (b) << (i)
    
    index_t r_Y = 0;		// rev(Y)
    for (unsigned i=0; i < l; i++) {
	// r_Y[l-1-i] = a_Y[i]
	SETBIT(r_Y, l-1-i, GETBIT(a,i));
    }

    return r_X | (r_Y << n-l);
}



void do_switch (vector<index_t> & V,
		index_t a, index_t b,
		bool should_swap)
{
    if (should_swap) {
	swap (V[a], V[b]);
    }
}


