/* -*- c++ -*-
 * alex iliev, feb 2003
 */

#ifndef _PERMUTE_H
#define _PERMUTE_H


#include <vector>

#include <common/utils.h>

#include "bit_matrix.h"
#include "lib.h"


// one index operation for a revslide permutation:
// <---n-l-----><--l--->
// |     X     |   Y   | goes to
// | rev(Y)|      X    |
// l is the level number
// n is the bit size of the indices
index_t revslide (index_t a, size_t l, size_t n);




// the _Switch object needs to have a method:
// operator () (index_t a, index_t b, bool should_switch)
template<typename _Switch>
void run_permutation_ct
(
    size_t N,
    const bit_matrix & S,	// the switches of the network
    _Switch & switch_fcn	// the switch function
    )
{
    
    unsigned g, k;
    size_t c;			// the column number

    const size_t n         = lgN_floor (N);
    const size_t n_sw      = lgN_floor (N/2); // lg (# of switches = N/2)
    const size_t col_count = 2*n-1;


    // c is the switch column
    for (c = 0; c < col_count; c++) {

	// l is the level number: 0, 1, .. n-2, n-1, .. 0
	unsigned l = (c < n)  ?  c  :  (col_count-1 - c);
	
	clog << "Benes level " << l << " @ " << print_time << endl;
	
	index_t s = 0;
	
	unsigned m = 1U << (l+1);
	
	bit_matrix::const_reference S_col = S[c];

	for (g = 0; g < N; g += m) {
	    // g iterates through the groups: it is the first wire in the
	    // current group

	    for (k = 0; k < m/2; k++) {
		// k is the switch number in the group

		switch_fcn (g+k, g+k+m/2, S_col[revslide (s, l, n_sw)]);

		s++;		// next switch

	    }

	}

    }

    
}


//
// callback functor to do a switch
//

class switch_functor {
public:
    void operator ()
	(std::vector<index_t> & V, index_t a, index_t b, bool should_swap);
};

typedef void (*switch_fcn_t)
    (std::vector<index_t> & V, index_t a, index_t b, bool should_swap);


//
// a permutation vector to carry both P and P'
//
struct perm_entry_t {
    // the positions change as follows under the permutation:
    index_t dest;		// i        -> P[i].dest
    index_t src;		// P[i].src -> i
};


typedef std::vector<perm_entry_t> perm_vector;


struct vector_stride_info_t {

    vector_stride_info_t () : start(0), stride(1) {}

    vector_stride_info_t (const vector_stride_info_t & a,
			  size_t start, size_t stride)
	: start (a.start + start * a.stride),
	  stride (a.stride * stride)
	{}
    

    size_t start, stride;
};


//
// index into a slice vector
//

template<class T>
typename std::vector<T>::reference
idx (std::vector<T> & V,
     const vector_stride_info_t & slice,
     typename std::vector<T>::size_type i)
{
    return V[slice.start + i * slice.stride];
}



template<class T>
typename std::vector<T>::const_reference
idx (const std::vector<T> & V,
     const vector_stride_info_t & slice,
     typename std::vector<T>::size_type i)
{
    return V[slice.start + i * slice.stride];
}



//
// generate a random permutation vector
//
void generate_permutation (perm_vector & V, size_t N);


// compute the switch network S for the permutation P
// 
// S should be the right size now, with all switches set to false, ie do not
// cross.
void
compute_switch_network
(
    const perm_vector & P, // the permutation
    size_t N,			// number of wires
    bit_matrix & switch_matrix,	// output parameter: the switches
    // the slice information for S
    const matrix_slice_info_t & slice = matrix_slice_info_t()
    );


#endif
