/*
 * ** PIR Private Directory Service prototype
 * ** Copyright (C) 2002 Alexander Iliev <iliev@nimbus.dartmouth.edu>
 * **
 * ** This program is free software; you can redistribute it and/or modify
 * ** it under the terms of the GNU General Public License as published by
 * ** the Free Software Foundation; either version 2 of the License, or
 * ** (at your option) any later version.
 * **
 * ** This program is distributed in the hope that it will be useful,
 * ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 * ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * ** GNU General Public License for more details.
 * **
 * ** You should have received a copy of the GNU General Public License
 * ** along with this program; if not, write to the Free Software
 * ** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * */

/*
 * pirsession.cc
 * alex iliev, nov 2002
 * implementation of the PIRSession class for handling retrievals
 */

#include <string>

#include <fstream>
#include <iostream>

#include <vector>
#include <algorithm>		// for swap()
#include <functional>		// for bind1st

#include <common/utils.h>
#include <common/record.h>
#include <common/hash.h>
#include <common/consts.h>

#include "lib.h"
#include "consts.h"
#include "pirsession.h"
#include "hostcall.h"
#include "io.h"


using namespace std;



#ifdef _TESTING_PIRSESSION


static const char RECORD_DELIM[] = "------";

static void pir_loop (const ByteBuffer & key, const ByteBuffer & mackey,
		      const vector<index_t>& shuffle)
    throw (host_exception, crypto_exception);



// TODO: the test driver is way out of date!
int main (int argc, char* argv[]) {

    //
    // read in all the records, encrypt them and write them out again
    // NOTE: pad them all to the same size. for now just figure out the max size
    // manually and allow some lax
    // eg for now: record name: 30 + 3 (for the size field)     = 33
    //             3 attributes x 2 (name,value) x 30 chars + 3 = 198
    //             about 250 bytes

    // todo: establish keys - one for encr and one for mac. save them to disk.
    // or rather use ones given on cmd line.
    //
    // read in the records one by one and write them out to the same positions
    // encrypted and macced.


    ByteBuffer keybuf, mackeybuf;
    

    if (argc < 4) {
	cerr << "Usage: " << argv[0] << " <op> <key file> <MAC key file>"
	     << endl;
	return 1;
    }

    
    string
	op = argv[1],
	keyfile = argv[2],
	mackeyfile = argv[3];

/*
    if (op == "genkeys") {
	genkeys (keyfile, mackeyfile);
	return EXIT_SUCCESS;
    }
*/
    
    loadkeys (keybuf, keyfile, mackeybuf, mackeyfile);



    vector<index_t> V;

    // run some retrievals
    
    load_shuffle (V, get_shuffle_filename());
    
    try {
	pir_loop (keybuf, mackeybuf, V);
    }
    catch (const exception& ex) {
	cerr << "Exception from pir_loop: " << ex.what() << endl;
	return EXIT_FAILURE;
    }
 
    return EXIT_SUCCESS;
}



static void pir_loop (const ByteBuffer & key, const ByteBuffer & mackey,
		      const vector<index_t>& shuffle)
    throw (host_exception, crypto_exception)
{


    PIRSession session (key, mackey, &shuffle, BUCKETSIZE);

    // and now go through a few retrievals ...

    string buf;
    while (getline (cin, buf)) {
	Record answer = session.find_record (buf);
	cout << answer << RECORD_DELIM << endl;
    }
    
}


#endif // _TESTING_PIRSESSION




PIRSession::PIRSession (const std::string& db_dir,
			SymCryptProvider & symprov,
			MacProvider & macprov,
			const ByteBuffer & key, const ByteBuffer & mackey,
			size_t bucket_size)
    throw (hostio_exception)
    :
    _sym_wrapper (key, symprov, mackey, macprov),
    _io          (db_dir),
    _bucket_size (bucket_size)
{
    // load up the shuffle vector
    ByteBuffer shuffle_file = _io.read (SHUFFLEFILE);

    istringstream is (std::string (shuffle_file.cdata(), shuffle_file.len()));

    for (index_t i; is >> i; ) {
	_shuffle.push_back (i);
    }
}



/*
 * retrieve a record with the given name.
 */
Record
PIRSession::find_record (const string& name)
    throw (host_exception, crypto_exception)
{

    Record the_rec;
    the_rec.name = "Dummy";	// give it some kind of name!

    bool already_fetched = false; // have we already fetched this record in
				  // this session?


    //
    // want to check whether there is a new shuffled DB available
    //
    // FIXME: quick hack to achieve this: read a file on the host
//    ifstream (
    index_t bucket_index = hash
	(reinterpret_cast<const unsigned char*>(name.c_str()), DBSIZE);
    
    // get the shuffled index
    index_t real_index = _shuffle[bucket_index];
    

    // go through all the previously fetched buckets/records and refetch them
    // from the host
    // keep the real one if it's already in there, and fetch a random one at the
    // end
    // if real one not in there, fetch *it* at the end

    for (vector<object_id>::const_iterator rid = _touched_records.begin();
 	 rid != _touched_records.end();
 	 rid++)
    {
	ByteBuffer enc_rec;

	// fetch the whole bucket, and keep the actual record bytes
	read_bucket (enc_rec, *rid);
    
	if (real_index == rid->bucket) {
	    // this may be the record we're looking for!
	    // could also be another record in the same bucket
	    Record r = decrypt_record (enc_rec);
	    if (r.name == name) {
		the_rec = r;
		already_fetched = true;
	    }
	}

    }
    

    object_id id;

    if (already_fetched) {
	// get a random record now
	Record temp;
	// FIXME: this is supposed to *not* be one we've fetched already!
	fetch_random_record(id, temp);
    }
    else {
	// get the actual one we need
	id.bucket = real_index;
	find_record_from_bucket (the_rec, id, name);
    }

    // save the id of the record we just fetched
    _touched_records.push_back (id);    

    
    return the_rec;
}


bool PIRSession::had_enough () throw () {
    // FIXME: apprently list::size() may be O(n) running time!
    return _touched_records.size() > 47;
}



// read through the whole bucket indicated by rid.bucket, and keep the the
// record rid.record from that bucket
void PIRSession::read_bucket (ByteBuffer & o_enc_rec,
			      const object_id& rid)
    throw (host_exception)
{
    ByteBuffer enc_rec;
    object_id id = rid;
    index_t rec_idx = id.record;

    for (index_t i=0; i < _bucket_size; i++) {
	id.record = i;
	enc_rec = _io.read (id);

	if (i == rec_idx) {
	    o_enc_rec = enc_rec;
	}
    }
}


	
#if 0
void
PIRSession::write_record (const ByteBuffer& rec, const RecordLocation& where) {

    //
    // need to:
    // pad, encrypt, and mac the record
    // send off the bytes to an appropriate host call to store them
    //

    // assume the rec is the right size. but how to pad it then?
    
    
}
#endif // 0


//
// examine all the records in this bucket, and store the correct one in o_rec
// gets the bucket number from id.bucket, then fills in the right id.record
//
void PIRSession::find_record_from_bucket (Record & o_rec,
					  object_id & id, const string& name)
    throw (crypto_exception, host_exception)
{

    index_t ours = 0;		// which record number is it?
    for (index_t r = 0; r < _bucket_size; r++) {
	
	Record current;

	id.record = r;
	fetch_record (current, id);

	// save if it's the right one
	if (current.name == name) {
	    // yes!!
	    o_rec = current;
	    ours = r;
	    // FIXME: this selective operation may leak timing information
	}

	// carry on...
    }

    id.record = ours;
}


//
// get a random record, returning both the record and its ID
//
void PIRSession::fetch_random_record (object_id & o_id,
				      Record & o_rec)
    throw (crypto_exception, host_exception) 
{
    o_id.bucket = random() % DBSIZE;
    o_id.record = random() % BUCKETSIZE;
	
    for (index_t r = 0; r < _bucket_size; r++) {
	
	Record current;
	object_id id = {o_id.bucket, r};
	fetch_record (current, id);

	// save if it's the right one
	if (r == o_id.record) {
	    // yes!!
	    o_rec = current;
	    // FIXME: this selective operation may leak timing information
	}

    }

}


void PIRSession::fetch_record (Record & o_rec, const object_id& id)
    throw (crypto_exception, host_exception)
{
    // get the ciphertext + mac from the host
    ByteBuffer enc_rec = _io.read (id);
    o_rec = decrypt_record (enc_rec);
}



#if 0
void
PIRSession::write_record (const ByteBuffer& rec, const RecordLocation& where) {

    //
    // need to:
    // pad, encrypt, and mac the record
    // send off the bytes to an appropriate host call to store them
    //

    // assume the rec is the right size. but how to pad it then?
    
    
}
#endif



Record PIRSession::decrypt_record (const ByteBuffer& hostbytes)
    throw (crypto_exception)
{
    return ::decrypt_record (hostbytes, _sym_wrapper);
}

