/*
 * ** PIR Private Directory Service prototype
 * ** Copyright (C) 2002 Alexander Iliev <iliev@nimbus.dartmouth.edu>
 * **
 * ** This program is free software; you can redistribute it and/or modify
 * ** it under the terms of the GNU General Public License as published by
 * ** the Free Software Foundation; either version 2 of the License, or
 * ** (at your option) any later version.
 * **
 * ** This program is distributed in the hope that it will be useful,
 * ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 * ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * ** GNU General Public License for more details.
 * **
 * ** You should have received a copy of the GNU General Public License
 * ** along with this program; if not, write to the Free Software
 * ** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * */

// shuffle.cc
// alex iliev jan 2003
// routines for PIR database shuffling, in the Shuffler class


#include <vector>
#include <fstream>
#include <list>
#include <functional>
#include <exception>

#include <unistd.h>		// for getopt()
#include <time.h>		// for time()
#include <signal.h>		// for SIGTRAP


#include <common/exceptions.h>
#include <common/record.h>
#include <common/sym_crypto.h>
#include <common/openssl_crypto.h>
#include <common/socket-class.h>
#include <common/consts.h>

#include "lib.h"
#include "hostcall.h"
#include "consts.h"
#include "permute.h"
#include "4758_sym_crypto.h"
#include "io.h"

#include "shuffle.h"



using namespace std;

static const std::string CLEARDIR = "clear",
    CRYPTDIR_BASE = "crypt";

static const unsigned short RETRIEVER_CARDNO = 2;


static void inform_retriever (const std::string & dbdir,
			      unsigned short retriever_cardnum)
    throw (comm_exception);



void out_of_memory () {
    cerr << "Out of memory!" << endl;
    raise (SIGTRAP);
    exit (EXIT_FAILURE);
}


static void
usage(const char* progname)
{
    fprintf(stderr,
	    "Usage: %s\n"
	    "[-p <card server port>]\n"
	    "[-d <root dir for keys etc.>]\n"
	    "[-c (means use 4758 crypto hardware instead of OpenSSL)]\n"
	    "[-r <retriever card number>]\n",
	    progname);
    exit(EXIT_FAILURE);
}



int main (int argc, char * argv[]) {


    unsigned short host_serv_port = PIR_HOSTSERV_PORT;
    unsigned short retriever_cardnum = RETRIEVER_CARDNO;
    string dir = ROOTDIR;
    bool use_card_crypt_hw = false; // do we use the 4758 hardware for crypto?
    
    
    ByteBuffer keybuf, mackeybuf;

    set_new_handler (out_of_memory);

    int opt;
    while ((opt = getopt(argc, argv, "p:d:r:c")) != EOF) {
	switch (opt) {

        case 'p':		// port
	    host_serv_port = strtol(optarg, NULL, 0);
	    if (host_serv_port == 0) {
		cerr << argv[0] << ": Invalid port" << endl;
		exit (EXIT_FAILURE);
	    }
	    break;

	case 'd':		//directory for keys etc
	    dir = optarg;
	    break;

	case 'c':
	    use_card_crypt_hw = true;
	    break;

	case 'r':
	    retriever_cardnum = atoi (optarg);
	    break;
	    
	default:
	    usage(argv[0]); 
	}
    }

    

    loadkeys (keybuf,    dir + DIRSEP + KEYFILE,
	      mackeybuf, dir + DIRSEP + MACKEYFILE);


    auto_ptr<SymCryptProvider> symprov;
    auto_ptr<MacProvider> macprov;

    try {
	if (use_card_crypt_hw) {
	    symprov.reset (new SymCrypt4758());
	    macprov.reset (new SHA1HMAC_4758());
	}
	else {
	    symprov.reset (new OSSLSymCrypto());
	    macprov.reset (new OSSL_HMAC());
	}
    }
    catch (const crypto_exception& ex) {
	cerr << "Exception from crypto provider creation!" << endl
	     << ex.what() << endl;
	return EXIT_FAILURE;
    }
    
    // DBSIZE and BUCKETSIZE are for now in lib.h

    for (int dbnum = 1; ; dbnum++) {

	std::ostringstream cryptdir_str;
	cryptdir_str << CRYPTDIR_BASE << dbnum;
	std::string cryptdir = cryptdir_str.str();
    
	Shuffler shuffler (cryptdir,
			   *symprov, *macprov,
			   keybuf, mackeybuf, DBSIZE, BUCKETSIZE);
    
	try {
	    shuffler.generate_shuffle();
	    shuffler.write_shuffle (SHUFFLEFILE);
	    shuffler.shuffle ();

	    // inform the retriever that a new DB is ready
	    inform_retriever (cryptdir, retriever_cardnum);
	}
	catch (const exception& ex) {
	    cerr << "Exception from shuffle: " << ex.what() << endl
		 << "Exiting" << endl;
	    return EXIT_FAILURE;
	}

    }

}


Shuffler::Shuffler (const std::string& crypt_dirname,
		    SymCryptProvider & symop, MacProvider & macop,
		    const ByteBuffer & key, const ByteBuffer & mackey,
		    size_t N, size_t bucksize)
    : _key         (key),
      _mackey      (mackey),
      _sym_wrapper (key, symop, mackey, macop),
      _crypt_io    (crypt_dirname),
      _clear_io    (CLEARDIR),
      N            (N),
      _bucket_size (bucksize)
{}



void Shuffler::generate_shuffle () {
    generate_permutation (V, N);
}




// funcion object to print an object with a '<<' operator defined
template<class T> struct print : public unary_function<T, void>
{
    void operator() (const T& x)
	{ os << x << endl; }

    print(ostream& out) : os(out) {}
    ostream& os;
};

// specialization of the above for perm_entry_t
template <> struct print<perm_entry_t> :
    public unary_function<perm_entry_t, void>
{
    // want to save the dest index for each src index
    void operator() (const perm_entry_t& x)
	{ os << x.dest << endl; }

    print(ostream& out) : os(out) {}
    ostream& os;
}; 




void Shuffler::write_shuffle (const string& filename)
    throw (std::ios::failure, hostio_exception)
{
    clog << "Writing shuffle vector to " << filename << endl;
    
    ostringstream oss;
    for_each (V.begin(), V.end(), print<perm_entry_t> (oss));

    _crypt_io.write (filename, ByteBuffer (oss.str()));
}    




void Shuffler::shuffle ()
    throw (host_exception, crypto_exception)
{
    clog << "Start encrypting DB @ " << print_time << endl;
    
    // should first encrypt all the records to produce the encrypted database
    make_encrypted_db();

    clog << "Done with encrypted DB @ " << print_time << endl;
    
    // allocate the switch settings bit matrix
    bit_vector S_template (N/2, false);
    bit_matrix S (2*lgN_floor(N) - 1, S_template);

    compute_switch_network (V, N, S);

    clog << "Network ready" << endl;
//    sleep (15);

    // perform the permutation using our own switch function object to perform
    // the actual switches
    Switcher switcher (*this);
    run_permutation_ct (N, S, switcher);

    clog << "Shuffle done @ " << print_time << endl;
}



void Shuffler::make_encrypted_db () throw (host_exception, crypto_exception)
{
    Bucket current;
    
    // read, encrypt and write all the buckets, no big deal

    for (index_t i=0; i < N; i++) {
	fetch_clear_bucket (i, current);

	write_bucket (i, current); // this encrypts too

	current.records.clear();
    }
}



void Shuffler::Switcher::operator () (index_t a, index_t b, bool should_swap)
    throw (host_exception, crypto_exception)
{

    // need to:
    // read the two buckets a and b
    // decrypt them
    // change the IV randomly (done by the SymDencrypter currently)
    // encrypt again
    // write them out, first into slot a, then b, swapped or not

    list<ByteBuffer> aphro, bachus;

//    clog << (should_swap ? "" : "no") << " swap " << a << "," << b << endl;
    

    // the blobs are also decrypted here
    master.fetch_blobs (a, aphro);
    master.fetch_blobs (b, bachus);

    list<ByteBuffer> *out1 = &aphro, *out2 = &bachus;
    if (should_swap) {
	swap (out1, out2);	// just swaps the pointers
				// FIXME: timing attack?
    }
	
    // always write the indices in the same order they were read. blobs are
    // re-encrypted using a different IV
    master.write_blobs (a, *out1);
    master.write_blobs (b, *out2);
}



//
// fetch an encrypted bucket and decrypt it into blobs. don't need the actual
// records here
//
void Shuffler::fetch_blobs (index_t n, list<ByteBuffer> & o_blobs)
    throw (host_exception, crypto_exception)
{
    ByteBuffer current_buf;
//    blob current_blob;

    for (index_t i = 0; i < _bucket_size; i++) {
	
	object_id id = {n, i};
	current_buf = _crypt_io.read (id);

	// decrypt
	o_blobs.push_back (_sym_wrapper.unwrap (current_buf));
		       
    }
	
}



//
// retrieve a whole bucket, in the clear.
//
void Shuffler::fetch_clear_bucket (index_t n, Bucket & o_bucket)
    throw (host_exception)
{

    // just fetch all the records in this buckets, ie until an error comes

    // FIXME: should somehow better be able to tell how many records in a
    // bucket, esp. since it should be the same across all buckets.

    for (index_t i = 0; i < _bucket_size; i++) {
//	clog << "Fetching bucket " << n << " record " << i << endl;
	
	object_id id = {n, i};
	ByteBuffer buf = _clear_io.read (id);
	Record current (buf);

// 	clog << "got a record! " << current.attributes->attr.name
// 	     << " = " << current.attributes->attr.value << endl;
	
	o_bucket.records.push_back (current);
    }
}



//
// write a bucket of records to the host, encrypting and maccing in the process
//

void Shuffler::write_bucket (index_t n, const Bucket & bucket)
    throw (crypto_exception, host_exception)
{

    // tactic: need to marshall, encrypt and mac each record, then write the
    // ciphertext out

    list<Record>::const_iterator it;
    int i = 0;

    for (it = bucket.records.begin();
	 it != bucket.records.end();
	 it++, i++)
    {
	object_id id = {n, i};
	ByteBuffer blob = encrypt_record(*it, _sym_wrapper);
	_crypt_io.write (id, blob);
    }
    
}


//
// write a list of cleartext blobs representing a bucket to the host
//
void Shuffler::write_blobs (index_t n, const list<ByteBuffer> & blobs)
    throw (crypto_exception, host_exception)
{

    list<ByteBuffer>::const_iterator it;
    int i = 0;

    for (it = blobs.begin(); it != blobs.end(); it++, i++) {
	object_id id = {n, i};
	ByteBuffer encblob = _sym_wrapper.wrap (*it);
	_crypt_io.write (id, encblob);
    }
}


void inform_retriever (const std::string & dbdir,
		       unsigned short retriever_cardnum)
    throw (comm_exception)
{

    SCCDatagramSocket sock;
    SCCSocketAddress retriever_addr (retriever_cardnum, PIR_CONTROL_PORT);

    sock.send (ByteBuffer (dbdir), retriever_addr);
}


// transpose is used a bit loosely here
// we just want to set V_[V[i]] = i for all indices i
void transpose_vector (vector<index_t> & V_, const vector<index_t>& V) {

    V_.resize (V.size());

    for (unsigned i = 0; i < V.size(); i++) {
	V_[V[i]] = i;
    }
}
